/*############################################################################
 Copyright:      Markus KOtschenreuther
 Author:         Markus KOtschenreuther
 Remarks:
 known Problems: none
 Version:        10.2011
 Description:
############################################################################*/
#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>

#include "config.h"
#include "std_defs.h"
#include "timer.h"
#include "uart.h"
#include "froeling_com.h"
#include "twislave.h"
#include "../common/twi_interface.h"
#include "../../../projShared/lanswitch.h"

//******************************************
// local variables
//******************************************
T_FROE_COM_DATA tFroeComStmData;

T_TWI_PARAM_TO_FROELINGSLAVE*	pParamFromMaster;
T_TWI_DAT_FROM_FROELINGSLAVE*	pDataToMaster;

//******************************************
//forward declarations
//******************************************
void Send_Msg(char* msg);
int  ComSend(char* SendBuffer, int bytes2send);
void CreateCommand(char* cmd, char* Param, int ParamLen, char* sendbuffer);
void CreateCheckSum(char* buf, int buffsize, char* checksum);
int  checkCheckSum(char* buf, int buffsize);
void sendAcq(char* buf, int buffsize);
void sendNAcq(char* buf, int buffsize);
int convert_FroelingTime(char x);
char convert_2_FroelingTime(char x);
void Set_Time(T_TWI_TIMESTRUCT_TO_FROELINGSLAVE* tTimeStruct);
void Set_ParamValue(unsigned long ParamID, unsigned long value);

//---------------------------------------
// functions to control statemachine
//---------------------------------------
void FroeCom_Init (void){

	// initialise state machine
	tFroeComStmData.tCurrentMainState = STM_FR_INIT;
	tFroeComStmData.tCurrentSubState = STM_SUB_FR_RECEIVE_HEADER;
};


//******************************************
//    STATEMACHINE
//******************************************
//---------------------------------------
int FroeComSTM(void)
{
	// Message Format:
	// Byte:	 1| 2| 3| 4| 5| 6  ...
	// --------------------------------------------
	// HEX:		4d|32|07|57|50|20|25|01|03|12|01|88
	// --------------------------------------------
	// ASCII:	 M| 2| ...
	// --------------------------------------------
	// Content: Ident| l|   Datablock        | CHS |
	// Blocks : Header  |   Body             | CHS |
	//
	// Byte 1,2:	Message identifier (M1 = measurement values / M2 = time)
	// Byte 3		Message data block length in bytes (l)
	// Byte 4...	Datablock
	// Last 2 Bytes:Checksum (CHS)

	// main statemachine
	// --------------------

	switch(tFroeComStmData.tCurrentMainState)
	{
		case STM_FR_INIT:
		{
			// init variables
			pParamFromMaster = (T_TWI_PARAM_TO_FROELINGSLAVE*)&rxbuffer[0];
			pDataToMaster = (T_TWI_DAT_FROM_FROELINGSLAVE*)&txbuffer[0];

			memset(pParamFromMaster, 0, sizeof(T_TWI_PARAM_TO_FROELINGSLAVE));
			memset(pDataToMaster, 0, sizeof(T_TWI_DAT_FROM_FROELINGSLAVE));

			memset(&tFroeComStmData,0, sizeof(T_FROE_COM_DATA));

			tFroeComStmData.glsUserID = FROELING_USER_ID;
			tFroeComStmData.ulLastSystemTimestamp = time;

			// init and reset uart
			uart_init();
			uart_flush();
			uart_fifo_flush();

			// start communication by sendig "Ra" to P3100
			Send_Msg("Ra");

			// next states
			tFroeComStmData.tCurrentSubState = STM_SUB_FR_RECEIVE_HEADER;
			tFroeComStmData.tCurrentMainState = STM_FR_RECEIVE_MSG;

			break;
		}

		case STM_FR_RECEIVE_MSG:
		{
			int iTmp = -1;

			// read one character from the input buffer (not blocking)
			iTmp = uart_getc_nowait();
			if (iTmp  > -1) {
				tFroeComStmData.cReceiveBuffer[tFroeComStmData.iInBuffPointer] = (unsigned char)iTmp;
				tFroeComStmData.iInBuffPointer++;
			}

			// sub statemachine
			// -----------------------
			switch(tFroeComStmData.tCurrentSubState)
			{
				case STM_SUB_FR_RECEIVE_HEADER:
				{
					if(tFroeComStmData.iInBuffPointer >= 3){
						// get parameter length
						tFroeComStmData.iMsgLength = (unsigned int)tFroeComStmData.cReceiveBuffer[2] + 5;
						if ((tFroeComStmData.iMsgLength) > RECEIVE_BUF_SIZE){
							tFroeComStmData.iInBuffPointer = 0;
							tFroeComStmData.iMsgLength = 0;
						}

						if (tFroeComStmData.cReceiveBuffer[0] == 'R'){
							// set next state
							tFroeComStmData.tCurrentSubState = STM_SUB_FR_RECEIVE_REPLY_BODY;
						}
						else if (tFroeComStmData.cReceiveBuffer[0] == 'M'){
							// set next state
							tFroeComStmData.tCurrentSubState = STM_SUB_FR_RECEIVE_MESSAGE_BODY;
						}
						else{
							tFroeComStmData.iInBuffPointer = 0;
						}
					}
					break;
				}

				case STM_SUB_FR_RECEIVE_REPLY_BODY:
				{
					// wait until 6 bytes have been received
					if(tFroeComStmData.iInBuffPointer >= 6){

						tFroeComStmData.iInBuffPointer = 0;
						tFroeComStmData.tCurrentSubState = STM_SUB_FR_RECEIVE_HEADER;
					}
					break;
				}
				case STM_SUB_FR_RECEIVE_MESSAGE_BODY:
				{
					// is telegram complete?
					if(tFroeComStmData.iInBuffPointer >= tFroeComStmData.iMsgLength){

						// check received checksum
						if(checkCheckSum(tFroeComStmData.cReceiveBuffer, tFroeComStmData.iMsgLength)){
							sendAcq(tFroeComStmData.cReceiveBuffer, tFroeComStmData.iMsgLength);
						}
						else{
							sendNAcq(tFroeComStmData.cReceiveBuffer, tFroeComStmData.iMsgLength);
						}

						// check if measurement values have been received (M1 messages)
						if(tFroeComStmData.cReceiveBuffer[1] == '1'){
							tFroeComStmData.tCurrentMainState = STM_FR_HANDLE_M1;
						}

						// check if date/time has been received
						else if(tFroeComStmData.cReceiveBuffer[1] == '2'){
							tFroeComStmData.tCurrentMainState = STM_FR_HANDLE_M2;
						}
						else
						{
							// next states
							tFroeComStmData.tCurrentSubState = STM_SUB_FR_RECEIVE_HEADER;
							tFroeComStmData.tCurrentMainState = STM_FR_RECEIVE_MSG;
						}
						tFroeComStmData.iInBuffPointer = 0;
					}
					break;
				}
			}
			break;
			// End of sub statemachine
			// -----------------------
		}
		case STM_FR_HANDLE_M1:
		{
			int i;
			int startindex = 3;

			for(i = 0; i < NB_OF_PROCESS_VALUES; i++){
				pDataToMaster->sProcValues[i] = (short)(((unsigned char)(tFroeComStmData.cReceiveBuffer[startindex + (i*2)]) << 8) + (unsigned char)tFroeComStmData.cReceiveBuffer[startindex + (i*2 + 1)]);
			}

			// next states
			tFroeComStmData.tCurrentSubState = STM_SUB_FR_RECEIVE_HEADER;
			tFroeComStmData.tCurrentMainState = STM_FR_RECEIVE_MSG;
			break;
		}
		case STM_FR_HANDLE_M2:
		{
			if(tFroeComStmData.cRbSendFlag == 0){
				// init measurement value transfer by sending "Rb" to P3100
				Send_Msg("Rb");
				tFroeComStmData.cRbSendFlag = 1;
			}

			// day
			pDataToMaster->ucTimeValues[0] = convert_FroelingTime(tFroeComStmData.cReceiveBuffer[6]);
			// month
			pDataToMaster->ucTimeValues[1] = convert_FroelingTime(tFroeComStmData.cReceiveBuffer[7]);
			// year
			pDataToMaster->ucTimeValues[2] = convert_FroelingTime(tFroeComStmData.cReceiveBuffer[9]);
			// hour
			pDataToMaster->ucTimeValues[4] = convert_FroelingTime(tFroeComStmData.cReceiveBuffer[4]);
			// min
			pDataToMaster->ucTimeValues[3] = convert_FroelingTime(tFroeComStmData.cReceiveBuffer[5]);
			// second
			pDataToMaster->ucTimeValues[5] = convert_FroelingTime(tFroeComStmData.cReceiveBuffer[3]);
			// Day of week
			pDataToMaster->ucTimeValues[6] = convert_FroelingTime(tFroeComStmData.cReceiveBuffer[8]);

			// next states
			tFroeComStmData.tCurrentSubState = STM_SUB_FR_RECEIVE_HEADER;
			tFroeComStmData.tCurrentMainState = STM_FR_RECEIVE_MSG;

			break;
		}
	}

	// communication watchdog
	if(time < tFroeComStmData.ulLastSystemTimestamp){
		// happends only at midnight...
		tFroeComStmData.ulLastSystemTimestamp = time;
	}
	else
	{
		if((time - tFroeComStmData.ulLastSystemTimestamp) >= ALIVE_CHECK_TIMESPAN){

			int iActFroelingTimestamp = (pDataToMaster->ucTimeValues[4] * 60) + pDataToMaster->ucTimeValues[5];

			// check if time sent by p3100 has changed since last check
			if(iActFroelingTimestamp == tFroeComStmData.iLastFroelingTimestamp){
				tFroeComStmData.tCurrentMainState = STM_FR_INIT;
			}
			tFroeComStmData.iLastFroelingTimestamp = iActFroelingTimestamp;
			tFroeComStmData.ulLastSystemTimestamp = time;
		}
	}

	//check whether we have to set a parameter value on P3100
	if((TransmissionActiveFlag == 0) && (pParamFromMaster->ulParamID > 0)){

		if(pParamFromMaster->ulParamID == FROE_PARAM_SETTIME){

			// recast pointer to correct format
			T_TWI_TIMESTRUCT_TO_FROELINGSLAVE* pTimeFromMaster;
			pTimeFromMaster = (T_TWI_TIMESTRUCT_TO_FROELINGSLAVE*)&rxbuffer;

			Set_Time(pTimeFromMaster);
		}
		else{
			// recast pointer to correct format
			T_TWI_LONGVALUE_TO_FROELINGSLAVE* pLongValFromMaster;
			pLongValFromMaster = (T_TWI_LONGVALUE_TO_FROELINGSLAVE*)&rxbuffer;

			Set_ParamValue(pLongValFromMaster->ulParamID, pLongValFromMaster->ulData);
		}

		// reset flag
		pParamFromMaster->ulParamID = 0;

		// quit action on master by sending back MsgId
		pDataToMaster->ucMsgID = pParamFromMaster->ucMsgID;
	}

	return(0);
}

/* ---------------------------------------------------
 * Construct and send a message in Frling-Format
 * ---------------------------------------------------
 */
void Send_Msg(char* msg)
{
	char sendbuffer[8];
	char arrayUserID[3];
	char b[8];
	char cs[2];

	arrayUserID[0] = 0;
	arrayUserID[1] = (char)((tFroeComStmData.glsUserID >> 8) & 0xFF);
	arrayUserID[2] = (char)(tFroeComStmData.glsUserID & 0xFF);

	memset(b, 0,(8)*sizeof(char));

	b[0] = msg[0];
	b[1] = msg[1];
	b[2] = 3;

	memcpy(&b[3], arrayUserID, 3);
	CreateCheckSum(b, 6, cs);
	memcpy(&b[6], cs, 2);
	memcpy(sendbuffer, b, 8);

	ComSend(sendbuffer, 8);
	return;
}

/* ---------------------------------------------------
 * Create a Frling Checksum
 * ---------------------------------------------------
 */
void CreateCheckSum(char* buf, int buffsize, char* checksum)
{
    unsigned short cs = 0;
    int i;

    for (i = 0; i < buffsize; i++){
        cs += (unsigned char)buf[i];
    }
    checksum[0] = (unsigned char)(cs >> 8);
    checksum[1] = (unsigned char) cs;
    return;
}

/* ---------------------------------------------------
 * Check if a checksum is correct
 * ---------------------------------------------------
 */
int checkCheckSum(char* buf, int buffsize)
{
	int i;
	int result = 0;
    unsigned short cs1 = 0;
    unsigned short cs2 = 0;

    // take last 2 bytes from received string (received checksum)
    cs1 = (unsigned short)(((unsigned char)buf[buffsize - 2] * 256) + (unsigned char)buf[buffsize - 1]);

    // build own checksum from received string
    for (i = 0; i < (buffsize - 2); i++){
        cs2 += (unsigned char)buf[i];
    }

    // compare both values
    if(cs1 == cs2){
    	result = 1;
    }
    return result;
}

/* ---------------------------------------------------
 * Send a Frling acknowledge
 * ---------------------------------------------------
 */
void sendAcq(char* buf, int buffsize)
{
	char b[6];
	char cs[2];

    if (buffsize > 1)
    {
        b[0] = buf[0];
        b[1] = buf[1];
        b[2] = 1;
        b[3] = 1;
        b[4] = 0;
        b[5] = 0;
        CreateCheckSum(b, 6, cs);
        memcpy(&b[4], cs, 2);
    }
    ComSend(b, 6);
    return;
}

/* ---------------------------------------------------
 * Send a "Not acknowledge" to Frling
 * ---------------------------------------------------
 */
void sendNAcq(char* buf, int buffsize)
{
	char b[6];
	char cs[2];

    if (buffsize > 1)
    {
        b[0] = buf[0];
        b[1] = buf[1];
        b[2] = 1;
        b[3] = 0;
        b[4] = 0;
        b[5] = 0;
        CreateCheckSum(b, 6, cs);
        memcpy(&b[4], cs, 2);
    }

    ComSend(b, 6);
    return;
}

/* ---------------------------------------------------
 * Send something via ComPort
 * ---------------------------------------------------
 */
int  ComSend(char* SendBuffer, int bytes2send)
{
	int i;

	for(i = 0; i < bytes2send; i++){
		uart_putc( SendBuffer[i]);
	}
	return(0);
}

/* ---------------------------------------------------
 * convert Frling Date/Time segment to an integer value
 * Frling interprets a Hex value directly "as it is written" as integer...
 * e.g. hour = 0x12 means hour = 12; minutes = 0x55 means minutes = 55
 * ---------------------------------------------------
 */
int convert_FroelingTime(char x)
{
	int y = 0;

	y = (int)(x>>4);
	y = (y*10) + (x-(y<<4));
	return(y);
}

char convert_2_FroelingTime(char x)
{
	char t = 0;
	char y = 0;

	t = ((int)(x/10));
	y = t*16 + (x-(t*10));
	return(y);
}

/* ---------------------------------------------------
 * Set time on P3100
 * ---------------------------------------------------
 */
void Set_Time(T_TWI_TIMESTRUCT_TO_FROELINGSLAVE* tTimeStruct)
{
	// message prototype
	//                   R     2    len   sec  min   hour  day  month  dow   year | Checksum
	//                   0     1     2     3     4     5     6     7     8     9
	char sendbuffer[]={0x52, 0x32, 0x07, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	char cs[2];

    sendbuffer[8] = convert_2_FroelingTime(tTimeStruct->cDayOfWeek);  // day of week
    sendbuffer[6] = convert_2_FroelingTime(tTimeStruct->cDay);  // day
    sendbuffer[7] = convert_2_FroelingTime(tTimeStruct->cMonth);  // month
    sendbuffer[9] = convert_2_FroelingTime(tTimeStruct->cYear);  // year
    sendbuffer[5] = convert_2_FroelingTime(tTimeStruct->cHour);  // hour
    sendbuffer[4] = convert_2_FroelingTime(tTimeStruct->cMin);  // min
    sendbuffer[3] = convert_2_FroelingTime(tTimeStruct->cSec);  // sec

	CreateCheckSum(sendbuffer, 10, cs);
	sendbuffer[10] = cs[0];
	sendbuffer[11] = cs[1];

	ComSend(sendbuffer, 12);

	return;
}

/* ---------------------------------------------------
 * Set a parameter on P3100
 * ---------------------------------------------------
 */
void Set_ParamValue(unsigned long ParamID, unsigned long value)
{

	// message prototype
	//                   R     I    len | ParamID   |  Value    | Checksum
	char sendbuffer[]={0x52, 0x49, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
	char cs[2];

	// set ParamID in msg prototype
	sendbuffer[3] = (char)((ParamID >> 8) & 0xFF);
	sendbuffer[4] = (char)(ParamID & 0xFF);

	// set mode in msg prototype
	sendbuffer[5] = (char)((value >> 8) & 0xFF);
	sendbuffer[6] = (char)(value & 0xFF);

	CreateCheckSum(sendbuffer, 7, cs);
	sendbuffer[7] = cs[0];
	sendbuffer[8] = cs[1];

	ComSend(sendbuffer, 9);

	return;
}
